@@ -29,6 +29,10 @@
 #' methods and has no effect if another method is chosen. Where an epsilon algorithm is used this should be one plus a multiple of two, ie (3,5,7,etc).
 #' @param Dampening This is the dampening parameter. By default it is 1 which means no dampening takes place. It can also be less than 1
 #' (indicating dampening) or more than 1 (indicating extrapolation).
+#' @param ReplaceInvalids This determines how NAs and Infs generated by the extrapolation method are handled. If it is "NoAction" then if
+#' the extrapolation algorithm generates an NA or Inf then no action is taken. This will likely result in these NAs and Infs being fed into
+#' the function and an error resulting. If it is "ReplaceVector" then the entire extrapolated vector will be replaced by a vector of a simple
+#' iterate. If it is "ReplaceElements" then the elements containing an NA or Inf will be replaced by the corresponding elements from a simple iterate.
 #' @param PrintReports This is a boolean describing whether to print ongoing ConvergenceMetric values for each iterate.
 #' @param ReportingSigFig This is the number of significant figures that will be used in printing the convergence values to the console
 #' (only if PrintReports is TRUE).
@@ -71,8 +75,9 @@
 #' F = FixedPoint(Function, Inputs, Method = "Anderson", MaxM = 4, ReportingSigFig = 13)
 FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "Simple", "Aitken", "Newton", "MPE", "RRE", "VEA", "SEA") ,
                       ConvergenceMetric  = function(Resids){max(abs(Resids))} , ConvergenceMetricThreshold = 1e-10, MaxIter = 1e3, MaxM = 10,
-                      ExtrapolationPeriod = 7, Dampening = 1, PrintReports = FALSE, ReportingSigFig = 5, ConditionNumberThreshold = 1e3,
-                      Plot = c("NoPlot", "ConvergenceFig", "ChangePerIterate"), ConvergenceFigLags = 5, ChangePerIteratexaxis = c()){
+                      ExtrapolationPeriod = 7, Dampening = 1, ReplaceInvalids = c("ReplaceElements","ReplaceVector","NoAction"), PrintReports = FALSE,
+                      ReportingSigFig = 5, ConditionNumberThreshold = 1e3, Plot = c("NoPlot", "ConvergenceFig", "ChangePerIterate"),
+                      ConvergenceFigLags = 5, ChangePerIteratexaxis = c()){
   # Checking inputs to the function
   Method = Method[1]
   Plot   = Plot[1]
@@ -128,18 +133,18 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
   Convergence = ConvergenceVector[iter]
   if (PrintReports){cat(paste0(format(" ", width = 49, justify = "right"), "Method: ", format(Method, width = 8, justify = "right")   ,
                                ". Iteration: ", format(iter, digits =  0, width = 5,scientific = FALSE),
-                               ". Convergence: ", NicePrint(Convergence, ReportingSigFig), "\n"))}
+                               ". Convergence: ",formatC(Convergence,  big.interval = 1, small.interval = ReportingSigFig, format = "e"),"\n"))}
   if (Plot == "ConvergenceFig"){ConvergenceFig(Inputs, Outputs,  Input_Convergence = ConvergenceVector, FromIterate = 1)}
   if (Plot == "ChangePerIterate"){ChangePerIterate(Inputs, Outputs, ConvergenceVector, FromIterate = dim(Inputs)[2],
                                                    ToIterate = dim(Inputs)[2], xaxis = ChangePerIteratexaxis, secondhold = -1)}
   iter = iter + 1
 
   while ((Convergence > ConvergenceMetricThreshold) & (iter <= MaxIter)){
-    # Generating new input and output.
+    # Generating new input guess
     NewInputVector = try({FixedPointNewInput(Inputs = Inputs, Outputs = Outputs, Method = Method, MaxM = MaxM,
                                               SimpleStartIndex = SimpleStartIndex, ExtrapolationPeriod = ExtrapolationPeriod,
                                               Dampening = Dampening, ConditionNumberThreshold = ConditionNumberThreshold,
-                                              PrintReports = PrintReports)})
+                                              PrintReports = PrintReports, ReplaceInvalids = ReplaceInvalids)})
     if (class(NewInputVector) == "try-error"){return(list(Inputs = Inputs, Outputs = Outputs, Convergence = ConvergenceVector,
                                                           FixedPoint = NA, Finish = "Could not generate new input vector"))}
     if (sum(is.na(NewInputVector))>0.5){return(list(Inputs = Inputs, Outputs = Outputs, Convergence = ConvergenceVector,
@@ -147,7 +152,7 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
     if (sum(is.infinite(NewInputVector))>0.5){return(list(Inputs = Inputs, Outputs = Outputs, Convergence = ConvergenceVector,
                                                     FixedPoint = NA, Finish = "New input vector contains Infs", NewInputVector = NewInputVector))}
     if (Method != "Anderson" & PrintReports){cat(paste0(format(" ", width = 49, justify = "right")))}
-
+    # Putting this input through the function to get a new output guess.
     NewOutputVector = try({Function(NewInputVector)})
     if (class(NewOutputVector) == "try-error"){return(list(Inputs = Inputs, Outputs = Outputs, Convergence = ConvergenceVector,
                                                            FixedPoint = NA, Finish = "Could not execute function with generated vector",
@@ -155,7 +160,7 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
     if (sum(is.na(NewOutputVector))){return(list(Inputs = Inputs, Outputs = Outputs, Convergence = ConvergenceVector,
                                                  FixedPoint = NA, Finish = "New output vector contains NAs", NewInputVector = NewInputVector,
                                                  NewOutputVector = NewOutputVector))}
-
+    # Adding input and output to saved matrices.
     Inputs  = matrix(c(Inputs, NewInputVector), ncol = iter, byrow = FALSE)
     Outputs = matrix(c(Outputs, NewOutputVector), ncol = iter, byrow = FALSE)
     Resid   = matrix(c(Resid, NewOutputVector - NewInputVector), ncol = iter, byrow = FALSE)
@@ -165,7 +170,7 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
     # Output of report and going to next iteration.
     if (PrintReports){cat(paste0("Method: ", format(Method, width = 8, justify = "right")   , ". Iteration: ",
                                  format(iter, digits =  0, width = 5,scientific = FALSE), ". Convergence: ",
-                                 NicePrint(Convergence, ReportingSigFig), "\n"))}
+                                 formatC(Convergence,  big.interval = 1, small.interval = ReportingSigFig, format = "e"), "\n"))}
     if (Plot == "ConvergenceFig"){ConvergenceFig(Inputs, Outputs,  Input_Convergence =  ConvergenceVector,
                                                  FromIterate = max(1, dim(Inputs)[2]- ConvergenceFigLags))}
     if (Plot == "ChangePerIterate"){ChangePerIterate(Inputs, Outputs, ConvergenceVector, secondhold = -1,
@@ -173,8 +178,9 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
                                                      xaxis = ChangePerIteratexaxis)}
     iter  = iter + 1
   }
+  # Extracting FixedPoint
   FixedPoint = Outputs[,dim(Outputs)[2]]
-
+  # Returning result
   if (Convergence < ConvergenceMetricThreshold){Finish = "Reached Convergence Threshold"} else {Finish = "Reached MaxIter"}
   return(list(Inputs = Inputs, Outputs = Outputs, Convergence = ConvergenceVector, FixedPoint = FixedPoint, Finish = Finish))
 }
@@ -198,6 +204,10 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
 #' @param ConditionNumberThreshold This is what threshold should be chosen to drop previous iterates if the matrix is ill conditioned.
 #' Only used in Anderson acceleration.
 #' @param PrintReports This is a boolean describing whether to print ongoing ConvergenceMetric values for each iterate.
+#' @param ReplaceInvalids This determines how NAs and Infs generated by the extrapolation method are handled. If it is "NoAction" then if
+#' the extrapolation algorithm generates an NA or Inf then no action is taken. This will likely result in these NAs and Infs being fed into
+#' the function and an error resulting. If it is "ReplaceVector" then the entire extrapolated vector will be replaced by a vector of a simple
+#' iterate. If it is "ReplaceElements" then the elements containing an NA or Inf will be replaced by the corresponding elements from a simple iterate.
 #' @return A vector containing the next guess in seeking a fixed point.
 #' @examples
 #' FPFunction = function(x){c(0.5*sqrt(abs(x[1] + x[2])), 1.5*x[1] + 0.5*x[2])}
@@ -207,32 +217,32 @@ FixedPoint = function(Function, Inputs, Outputs = c(), Method = c("Anderson", "S
 #' NewGuessMPE = FixedPointNewInput(A$Inputs, A$Outputs, Method = "MPE")
 #' NewGuessAitken = FixedPointNewInput(A$Inputs, A$Outputs, Method = "Aitken")
 FixedPointNewInput = function(Inputs, Outputs, Method = "Anderson", MaxM = 10, SimpleStartIndex = 1,
-                               ExtrapolationPeriod = 7, Dampening = 1, ConditionNumberThreshold = 1e3, PrintReports = FALSE ) {
-
+                               ExtrapolationPeriod = 7, Dampening = 1, ConditionNumberThreshold = 1e3,
+                              PrintReports = FALSE, ReplaceInvalids = c("ReplaceElements","ReplaceVector","NoAction") ) {
   CompletedIters = dim(Outputs)[2]
+  ReplaceInvalids = ReplaceInvalids[1]
 
   if (Method == "Simple"){
     return( as.vector(Outputs[,CompletedIters] ))
-    }
-  if (Method == "Anderson"){
+  } else if (Method == "Anderson"){
     if (CompletedIters < 1.5){
       if (PrintReports){cat(paste0(format(" ", width = 32, justify = "right"), "  Using",  format(0, width = 3, justify = "right")," lags. "))}
       return(as.vector( Outputs[,CompletedIters] ))
-      }
+    }
+    # Extract variables for anderson acceleration algorithm.
     VectorLength   = dim(Outputs)[1]
     M = min(MaxM-1,CompletedIters-1,VectorLength)
     Coeffs = NA
-
-    Outputs         = matrix(Outputs[, seq(CompletedIters-M, CompletedIters,1)], ncol = M+1)
-    Inputs          = matrix(Inputs[ , seq(CompletedIters-M, CompletedIters,1)], ncol = M+1)
-    Resid           = Outputs - Inputs
-    DeltaOutputs    = matrix(Outputs[,2:(M+1)] - Outputs[,1:M], ncol = M)
-    DeltaResids     = matrix(Resid[  ,2:(M+1)] - Resid[  ,1:M], ncol = M)
-
+    Truncated_Outputs = matrix(Outputs[, seq(CompletedIters-M, CompletedIters,1)], ncol = M+1)
+    Truncated_Inputs  = matrix(Inputs[ , seq(CompletedIters-M, CompletedIters,1)], ncol = M+1)
+    Resid             = Truncated_Outputs - Truncated_Inputs
+    DeltaOutputs      = matrix(Truncated_Outputs[,2:(M+1)] - Truncated_Outputs[,1:M], ncol = M)
+    DeltaResids       = matrix(Resid[  ,2:(M+1)] - Resid[  ,1:M], ncol = M)
     LastResid       = Resid[,M+1]
-    LastOutput      = Outputs[,M+1]
-
+    LastOutput      = Truncated_Outputs[,M+1]
+    # We nest the fitting of coefficients in a while loop. This allows us to drop columns and try again if we get NAs from a lack of numerical stability.
     while (sum(is.na(Coeffs))>0.5){
+      # For numerical stability reasons we monitor the condition number
       ConditionNumber = rcond(DeltaResids)
       if (ConditionNumber > ConditionNumberThreshold){
         M = M-1
@@ -240,28 +250,28 @@ FixedPointNewInput = function(Inputs, Outputs, Method = "Anderson", MaxM = 10, S
         DeltaResids = matrix(DeltaResids[,2:(M+1)], ncol = M)
         next
       }
-
+      # We fit a linear model to get coefficients.
       Fit             = stats::lm.fit(x = DeltaResids, y = LastResid, tol = 1e-13)
       Coeffs          = Fit$coefficients
-
+      # If we have NAs then we drop a previous iterate and will try again in the next loop.
       if (sum(is.na(Coeffs))>0.5){
       M = M-1
       if (M < 1.5){
         # This happens occasionally in test cases where the iteration is very close to a fixed point.
         if (PrintReports){cat(paste0(format(" ", width = 32, justify = "right"), "  Using",
                                      format(0, width = 3, justify = "right")," lags. "))}
-        return( LastOutput )
+        return(LastOutput)
       }
       DeltaOutputs= matrix(DeltaOutputs[, 2:(M+1)], ncol = M)
       DeltaResids = matrix(DeltaResids[,2:(M+1)], ncol = M)
       }
     }
+    # Print information on the use of the anderson method and return the new guess.
     if (PrintReports){cat(paste0("Condition number is ", format(format(ConditionNumber, digits = 5,scientific = TRUE), width = 12, justify = "right")
                                  ,". Used:",  format(M+1, width = 3, justify = "right"), " lags. "))}
     NewGuess        = LastOutput - Dampening * Coeffs  %*% t(DeltaOutputs)
-    return( as.vector(NewGuess) )
-  }
-  if (Method == "Aitken"){
+    NewGuess = as.vector(NewGuess)
+  } else if (Method == "Aitken"){
     if ((CompletedIters + SimpleStartIndex) %% 3 == 0){
       # If we are in 3rd, 6th, 9th, 12th iterate from when we started Acceleration then we want to do a jumped Iterate,
       # First we extract the guess that started this run of 3 iterates (x), the Function applied to it (fx) and the function applied to that (ffx)
@@ -270,19 +280,13 @@ FixedPointNewInput = function(Inputs, Outputs, Method = "Anderson", MaxM = 10, S
       ffx = Outputs[,CompletedIters ]
       # Now using the appropriate formula to make a new guess. Note that if a vector is input here it is used elementwise.
       NewGuess = x - ((fx - x)^2/(ffx - 2*fx + x))
-      # Now there is the chance that the demoninator is zero which results in an Inf.
-      if (sum(is.infinite(NewGuess)) + sum(is.na(NewGuess)) > 0.5){
-         NewGuess[(is.infinite(NewGuess) | is.na(NewGuess))] = Outputs[(is.infinite(NewGuess) | is.na(NewGuess)) ,CompletedIters]
-        }
-      return( as.vector(Dampening*NewGuess + (1-Dampening)*Outputs[,CompletedIters] ))
     } else {
       # We just do a simple iterate. We do an attempt with the latest iterate.
       return( as.vector(Outputs[,CompletedIters] ))
       }
-  }
-  if (Method == "Newton"){
+  } else if (Method == "Newton"){
     if ((CompletedIters + SimpleStartIndex) %% 2 == 1 & CompletedIters > 1){
-      # If we are in 3rd, 6th, 9th, 12th iterate from when we started Newton Acceleration then we want to do a Newton Iterate,
+      # If we are in 3rd, 6th, 9th, 12th iterate from when we started Acceleration then we want to do a jumped Iterate,
       # First we extract the guess that started this run of 3 iterates (xk1), the Function applied to it (fxk1, xk) and the function applied to that (fxk)
       xk1 = Inputs[,CompletedIters -1]
       fxk1 = Outputs[,CompletedIters -1]
@@ -293,36 +297,39 @@ FixedPointNewInput = function(Inputs, Outputs, Method = "Anderson", MaxM = 10, S
       # Now using the appropriate formula to make a new guess. Note that if a vector is input here it is used elementwise.
       derivative = (gxk-gxk1)/(xk - xk1)
       NewGuess = xk - (gxk /derivative)
-      # Now checking and replacing elements where the derivative was zero.
-      if (sum(is.infinite(NewGuess)) + sum(is.na(NewGuess)) > 0.5){
-        NewGuess[(is.infinite(derivative) | is.na(derivative))] = Outputs[(is.infinite(derivative) | is.na(derivative)) ,CompletedIters]
-      }
-      return(as.vector(Dampening*NewGuess + (1-Dampening)*Outputs[,CompletedIters] ))
     } else {
-      # We just do a simple iterate.
-      return(as.vector( Outputs[,CompletedIters] ))
+      # We just do a simple iterate. We do an attempt with the latest iterate.
+      return( as.vector(Outputs[,CompletedIters] ))
     }
-  }
-  if (Method == "MPE" | Method == "RRE"){
+  } else if (Method == "MPE" | Method == "RRE"){
     SimpleIteratesMatrix = PutTogetherIteratesWithoutJumps(Inputs, Outputs)
     if ( dim(SimpleIteratesMatrix)[2] %% ExtrapolationPeriod == 0  ){
-        NewGuess = PolynomialExtrapolation(SimpleIteratesMatrix, Method = Method)
-        return(as.vector(Dampening*NewGuess + (1-Dampening)*Outputs[,CompletedIters]))
+      NewGuess = PolynomialExtrapolation(SimpleIteratesMatrix, Method = Method)
     } else {
       # We just do a simple iterate.
-      return( as.vector(Outputs[,CompletedIters]) )
+      return(as.vector(Outputs[,CompletedIters]))
     }
-  }
-  if (Method == "VEA" | Method == "SEA"){
+  } else if (Method == "VEA" | Method == "SEA"){
     SimpleIteratesMatrix = PutTogetherIteratesWithoutJumps(Inputs, Outputs)
     if ( dim(SimpleIteratesMatrix)[2] %% ExtrapolationPeriod == 0  ){
       NewGuess = EpsilonExtrapolation(SimpleIteratesMatrix, Method = Method)
-      return(as.vector(Dampening*NewGuess + (1-Dampening)*Outputs[,CompletedIters]))
     } else {
       # We just do a simple iterate.
-      return( as.vector(Outputs[,CompletedIters] ))
+      return(as.vector(Outputs[,CompletedIters] ))
     }
   }
+  # Here we use the ReplaceInvalids input to remove NAs and Infs.
+  if (ReplaceInvalids == "ReplaceElements"){
+    if (sum(is.infinite(NewGuess)) + sum(is.na(NewGuess)) > 0.5){
+      NewGuess[(is.infinite(NewGuess) | is.na(NewGuess))] = Outputs[(is.infinite(NewGuess) | is.na(NewGuess)) ,CompletedIters]
+  } else if (ReplaceInvalids == "ReplaceVector"){
+    if (sum(is.infinite(NewGuess)) + sum(is.na(NewGuess)) > 0.5){
+      NewGuess = Outputs[,CompletedIters]
+     }
+  }
+  # Here we apply any dampening and return the new guess.
+  return(as.vector(Dampening*NewGuess + (1-Dampening)*Outputs[,CompletedIters] ))
+  }
 }
 
 
@@ -353,8 +360,8 @@ PolynomialExtrapolation = function(Iterates, Method = c("MPE", "RRE")){
                                    ncol = TotalColumnsOfIterates-2)
     LastDifference       = matrix(  (Iterates[,TotalColumnsOfIterates]-Iterates[,(TotalColumnsOfIterates-1)])     , ncol = 1)
     InverseDifferences   = MASS::ginv(Differences)
-    cVector               = -InverseDifferences %*% LastDifference
-    cVector               = rbind(cVector,1)
+    cVector              = -InverseDifferences %*% LastDifference
+    cVector              = rbind(cVector,1)
     s                    = (Iterates[,2:TotalColumnsOfIterates]%*%cVector)/sum(cVector)
     return(s)
   }
@@ -428,8 +435,6 @@ EpsilonExtrapolationVectorOfInverses = function(DifferenceMatrix, Method){
     }
 }
 
-
-
 #' PutTogetherIteratesWithoutJumps
 #' This function takes the previous inputs and outputs and assembles a matrix with both excluding jumps.
 #' @export
@@ -448,33 +453,22 @@ EpsilonExtrapolationVectorOfInverses = function(DifferenceMatrix, Method){
 #'A = FixedPoint( function(x){c(0.5*sqrt(abs(x[1] + x[2])), 1.5*x[1] + 0.5*x[2])},
 #' Inputs = A$Inputs, Outputs = A$Outputs, MaxIter = 5, Method = "Simple")
 #'CombinedSequence = PutTogetherIteratesWithoutJumps(A$Inputs, A$Outputs)
-PutTogetherIteratesWithoutJumps = function(Inputs, Outputs, AgreementThreshold = 1e-10){
+PutTogetherIteratesWithoutJumps = function(Inputs, Outputs, AgreementThreshold = 10 * .Machine$double.eps){
+  # Input checking
   if (!(is.matrix(Inputs))){stop("Inputs is not a matrix.")}
   if (!(is.matrix(Outputs))){stop("Outputs is not a matrix.")}
   if (any(dim(Inputs) != dim(Outputs))){stop("Inputs and Outputs matrices are not comformable.")}
   Dimensions = dim(Inputs)
-
+  # We get a matrix of the differences between outputs and the next inputs. If they were all simple iterates these should match.
   if (Dimensions[2] == 1){return(matrix(cbind(matrix(Inputs, ncol = 1), matrix(Outputs, ncol = 1)), ncol =2))}
   Difference = matrix(Inputs[,2:(Dimensions[2])] - Outputs[,1:(Dimensions[2]-1)], nrow = Dimensions[1])
-
+  # We add up differences to get which iterates were simple iterates.
   Sum_Of_Differences = apply(Difference, 2, function(x) {sum(abs(x))})
   Agreements = Sum_Of_Differences < AgreementThreshold
   if (all(Agreements)){return(cbind(matrix( Inputs[,1:(Dimensions[2])], nrow=Dimensions[1]), matrix(Outputs[,Dimensions[2]],nrow=Dimensions[1])))}
-
+  # If not all iterates were simple we just take the latter iterates and return them.
   LocationsOfBreaks = which(!Agreements)
   LastBreak = LocationsOfBreaks[length(LocationsOfBreaks)]
   return(matrix(cbind(matrix(Inputs[,(LastBreak+1):(Dimensions[2])], ncol = Dimensions[2] - LastBreak),
                       matrix(Outputs[, Dimensions[2]], ncol = 1)), ncol = Dimensions[2] - LastBreak + 1))
 }
-
-
-#' NicePrint
-#'
-#' This is a small function to ensure that decimal places line up for convergence values reported by the fixedpoint function.
-#' @param Number The number to be formatted.
-#' @param sigfig The number of significant figures to be taken.
-#' @return A nicely formatted string version of the input number for printing to the console.
-NicePrint = function(Number, sigfig = 5){
-  CPrint    = formatC(Number,  big.interval = 1, small.interval = sigfig, format = "e")
-  return(CPrint)
-}
